@@ -32,6 +32,7 @@ G_BEGIN_DECLS
 #define VINAGRE_IS_TAB(obj)           (G_TYPE_CHECK_INSTANCE_TYPE((obj), VINAGRE_TYPE_TAB))
 #define VINAGRE_IS_TAB_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), VINAGRE_TYPE_TAB))
 #define VINAGRE_TAB_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS((obj), VINAGRE_TYPE_TAB, VinagreTabClass))
+#define VINAGRE_TAB_KEY               "VINAGRE_TAB_KEY"
 
 typedef struct _VinagreTabPrivate VinagreTabPrivate;
 typedef struct _VinagreTab        VinagreTab;
@@ -84,6 +85,7 @@ void		  vinagre_tab_set_notebook	(VinagreTab *tab,
 VinagreNotebook   *vinagre_tab_get_notebook	(VinagreTab *tab);
 
 void		  vinagre_tab_take_screenshot	(VinagreTab *tab);
+void		  vinagre_tab_send_ctrlaltdel	(VinagreTab *tab);
 void		  vinagre_tab_paste_text	(VinagreTab *tab,
 						 const gchar *text);
 
@@ -92,9 +94,15 @@ gboolean	  vinagre_tab_get_scaling	(VinagreTab *tab);
 void	          vinagre_tab_set_readonly	(VinagreTab *tab, gboolean active);
 gboolean	  vinagre_tab_get_readonly	(VinagreTab *tab);
 
-VinagreTabState   vinagre_tab_get_state		(VinagreTab *tab);
+gint		  vinagre_tab_get_original_width  (VinagreTab *tab);
+gint		  vinagre_tab_get_original_height (VinagreTab *tab);
+void		  vinagre_tab_original_size	  (VinagreTab *tab);
 
+VinagreTabState   vinagre_tab_get_state		(VinagreTab *tab);
+VinagreTab	  *vinagre_tab_get_from_connection (VinagreConnection *conn);
+
+gboolean	  vinagre_tab_is_pointer_grab	(VinagreTab *tab);
 G_END_DECLS
 
 #endif  /* __VINAGRE_TAB_H__  */
-/* vim: ts=8 */
+/* vim: set ts=8: */
